#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "macros.h"

#define RIGHT_CHILD(i)  (((i) * 2) + 2)
#define LEFT_CHILD(i)   (((i) * 2) + 1) 
#define PARENT(i)       (((i) - 1) / 2)
#define INIT_SIZE   1
#define GROW_SIZE	32

#define CONTAINER_OF(ptr, type, field) \
	((type *)((unsigned long)(ptr) - (unsigned long)(&((type *)(0))->field)))

typedef int (*prioq_compare_t)(struct prioq_head *a, struct prioq_head *b);

struct prioq_head {
	int index;
};		

struct prioq {
	struct prioq_head   **heap;
	size_t              size;
	size_t              used;
	prioq_compare_t     compare;
};	



struct prioq *
prioq_new(prioq_compare_t compare)
{
	struct prioq *queue;

	queue = (struct prioq*)malloc(sizeof(struct prioq));
	
	return_val_if_fail(queue !=NULL, NULL);
	
	memset(queue, 0, sizeof(*queue));

	queue->heap = (struct prioq_head**)malloc(sizeof(struct prioq_head *) *INIT_SIZE);

	if (queue->heap == NULL) {
			fprintf(stderr, "prioq_new: can't allocate heap\n");
			free(queue);
			return NULL;
	}

	memset(queue->heap, 0, sizeof(struct prioq_head *) * INIT_SIZE);

	queue->size    = INIT_SIZE;
	queue->used    = 0;
	queue->compare = ((compare != NULL) ? compare : compare);
	return queue;
}

#define NEW_SIZE(q)	(((q)->size + GROW_SIZE) * sizeof(struct prioq_head *))
#define OLD_SIZE(q)	((q)->size * sizeof(struct prioq_head *))

int 
prioq_put(struct prioq *queue, struct prioq_head *hp)
{
	int idx;
	struct prioq_head** tmp;
	
	return_val_if_fail(queue != NULL && hp != NULL,NULL);	
	
	if (queue->size == queue->used) {
		tmp = (struct prioq_head**)realloc(queue->heap, NEW_SIZE(queue));
		        
		if (!tmp) {
			fprintf(stderr, "prioq_put: can't realloc heap\n");
			return -1;
		}
		queue->heap = tmp;
		
		memset(queue->heap + queue->size, 0,
		NEW_SIZE(queue) - OLD_SIZE(queue));
				
		queue->size += GROW_SIZE;
	}

	idx = queue->used++;
	while (idx > 0) {
		if (queue->compare(hp, queue->heap[PARENT(idx)]) < 0)
			break;
		queue->heap[idx]        = queue->heap[PARENT(idx)];
		queue->heap[idx]->index = idx;
		idx = PARENT(idx);
    }
     
    queue->heap[idx]        = hp;
    queue->heap[idx]->index = idx;	
	
	return 0;
}	


static int
compare(struct prioq_head *a, struct prioq_head *b)
{
	struct user_data *ap = (struct user_data *) a;
	struct user_data *bp = (struct user_data *) b;

	if (ap->data > bp->data)
		return 1;
	else if (ap->data < bp->data)
		return -1;
	return 0;
}


static void
prioq_heapify(struct prioq *queue, int idx)
{
	unsigned int idx_max;
    unsigned int idx_right, idx_left;	
	struct prioq_head*  tmp;
	
	while ((idx_left = LEFT_CHILD(idx)) < queue->used) {
		idx_right = RIGHT_CHILD(idx);
		
		if ((idx_right < queue->used) && (queue->compare(queue->heap[idx_right],
			queue->heap[idx_left]) > 0))
			idx_max = idx_right;
		else 
            idx_max = idx_left;
		if (queue->compare(queue->heap[idx], queue->heap[idx_max]) < 0) {
			tmp = queue->heap[idx];
			queue->heap[idx]        = queue->heap[idx_max];
			queue->heap[idx]->index = idx;
			queue->heap[idx_max]    = tmp;
			queue->heap[idx_max]->index = idx_max;
			idx = idx_max;
		} else
			break;
	}
}



void
prioq_remove(struct prioq *queue, struct prioq_head *hp) 
{
    int idx;
    
	return_if_fail(queue != NULL && hp != NULL);
	
	idx       = hp->index;
    hp->index = -1;
		
	if (hp->index > (int)queue->used)
		fprintf(stderr,"prioq_remove: index out of range\n");
	
	queue->used--;
	
	if ((queue->used > 0) && (idx != (int)queue->used)) {
		
		while (idx > 0) {
			
			if (queue->compare(queue->heap[queue->used], queue->heap[PARENT(idx)]) < 0)
				break;
			
			queue->heap[idx] = queue->heap[PARENT(idx)];
			queue->heap[idx]->index = idx; 
			idx = PARENT(idx);
		}
		queue->heap[idx] = queue->heap[queue->used];
		queue->heap[idx]->index = idx;
	
		prioq_heapify(queue,idx);
	}

}	

void
prioq_free(struct prioq** queue)
{
	if (*queue == NULL) {
		fprintf(stderr,"prioq_free: The queue is free\n");
		return;
	}
	if ((*queue)->heap == NULL) {
		free(*queue);
		fprintf(stderr, "prioq_free: The queue id free\n");
		return;
	}
	
	free((*queue)->heap);
	free(*queue);
	*queue = NULL;
	fprintf(stderr, "prioq_free: The queue is free\n");
}	
	
	
   
	

struct prioq_head *
prioq_peek(struct prioq* queue) 
{
	struct prioq_head *head;
	struct user_data  *ptr;
	
	return_val_if_fail(queue != NULL, NULL);
			
	if (queue->used == 0) 
		return NULL;		
	
	head        = queue->heap[0];
	head->index = 0;
	ptr         = (struct user_data*)head;
	fprintf(stderr, "prioq_peek: data %d index %d\n",
			             ptr->data, ptr->head.index);
	return head;
}

struct prioq_head *
prioq_get(struct prioq *queue)
{
	struct prioq_head *head;
	struct user_data  *ptr;
	
	return_val_if_fail(queue != NULL,NULL)	    
		
	if (queue->used == 0)
		return NULL;		
 	
	head = queue->heap[0];
	head->index = 0;
	ptr  = (struct user_data*)head;
	queue->used--;
	
	fprintf(stderr,"prioq_get: data %d\n", ptr->data);
	
	if (queue->used > 0) {
		queue->heap[0]        = queue->heap[queue->used];
		queue->heap[0]->index = 0;
		prioq_heapify(queue, 0);
	}
	
	return head;	
}	